from inflection import singularize

from .operation import Operation
from quandl.connection import Connection
from quandl.util import Util


class GetOperation(Operation):

    @classmethod
    def get_path(cls):
        return cls.default_path()

    def __get_raw_data__(self):
        if self._raw_data:
            return self._raw_data

        cls = self.__class__
        params = {'id': str(self.code)}
        options = Util.merge_options('params', params, **self.options)

        path = Util.constructed_path(cls.get_path(), options['params'])

        r = Connection.request('get', path, **options)
        response_data = r.json()
        Util.convert_to_dates(response_data)
        self._raw_data = response_data[singularize(cls.lookup_key())]
        return self._raw_data



from quandl.errors.quandl_error import InvalidRequestError
from .utils.api_key_util import ApiKeyUtil
from .model.datatable import Datatable
from .message import Message


def export_table(datatable_code, **kwargs):
    """Downloads an entire table as a zip file.
    :param str datatable_code: The datatable code to download, such as MER/F1
    :param str filename: The filename for the download. \
    If not specified, will download to the current working directory
    :param str api_key: Most databases require api_key for bulk download
    """

    # discourage users from using authtoken
    if 'authtoken' in kwargs:
        raise InvalidRequestError(Message.ERROR_AUTHTOKEN_NOT_SUPPORTED)

    ApiKeyUtil.init_api_key_from_args(kwargs)

    filename = kwargs.pop('filename', '.')
    return Datatable(datatable_code).download_file(filename, **kwargs)



import re
import unittest
from test.helpers.httpretty_extension import httpretty
import json
import datetime
import six
from quandl.model.dataset import Dataset
from quandl.model.database import Database
from mock import patch, call
from test.factories.dataset import DatasetFactory
from test.factories.meta import MetaFactory


class GetDatasetTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        httpretty.enable()
        dataset = {'dataset': DatasetFactory.build(
            database_code='NSE', dataset_code='OIL')}
        httpretty.register_uri(httpretty.GET,
                               re.compile(
                                   'https://www.quandl.com/api/v3/datasets/*'),
                               body=json.dumps(dataset))
        dataset_code = Dataset.get_code_from_meta(dataset['dataset'])
        cls.dataset_instance = Dataset(dataset_code, dataset['dataset'])

    @classmethod
    def tearDownClass(cls):
        httpretty.disable()
        httpretty.reset()

    @patch('quandl.connection.Connection.request')
    def test_dataset_calls_connection(self, mock):
        d = Dataset('NSE/OIL')
        d.data_fields()
        expected = call('get', 'datasets/NSE/OIL/metadata', params={})
        self.assertEqual(mock.call_args, expected)

    def test_dataset_returns_dataset_object(self):
        dataset = Dataset('NSE/OIL')
        self.assertIsInstance(dataset, Dataset)
        self.assertEqual(dataset.dataset_code, 'OIL')

    def test_dataset_attributes_are_datetime_objects(self):
        dataset = Dataset('NSE/OIL')
        self.assertIsInstance(dataset.refreshed_at, datetime.datetime)
        self.assertIsInstance(dataset.newest_available_date, datetime.date)

    def test_dataset_column_names_match_expected(self):
        dataset = Dataset('NSE/OIL')
        six.assertCountEqual(self,
                             dataset.column_names, ['Date', 'column.1', 'column.2', 'column.3'])

    def test_dataset_database_gives_instance_of_database(self):
        database = self.dataset_instance.database()
        self.assertIsInstance(database, Database)

    @patch('quandl.model.data.Data.all')
    def test_dataset_data_calls_data_all(self, mock):
        self.dataset_instance.data(
            params={'start_date': '2015-07-01', 'end_date': '2015-07-10'})
        expected = call(params={'dataset_code': 'OIL', 'database_code': 'NSE',
                                'start_date': '2015-07-01', 'end_date': '2015-07-10',
                                'order': 'asc'})
        self.assertEqual(mock.call_args, expected)


class ListDatasetsTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        httpretty.enable()
        datasets = {'datasets': DatasetFactory.build_batch(3)}
        meta = {'meta': MetaFactory.build(current_page=1, total_pages=1)}
        datasets.update(meta)
        httpretty.register_uri(httpretty.GET,
                               re.compile(
                                   'https://www.quandl.com/api/v3/datasets*'),
                               body=json.dumps(datasets))

    @classmethod
    def tearDownClass(cls):
        httpretty.disable()
        httpretty.reset()

    @patch('quandl.connection.Connection.request')
    def test_datasets_calls_connection(self, mock):
        Dataset.all()
        expected = call('get', 'datasets', params={})
        self.assertEqual(mock.call_args, expected)

    def test_datasets_return_dataset_objects(self):
        results = Dataset.all()
        self.assertEqual(len(results), 3)
        for result in results:
            self.assertIsInstance(result, Dataset)

    def test_datasets_has_more_results(self):
        results = Dataset.all()
        self.assertFalse(results.has_more_results())

    def test_values_exist(self):
        results = Dataset.all()
        self.assertIsInstance(results.values, list)
        self.assertIsInstance(results.values[0], Dataset)

    def test_meta_attributes_can_be_accessed(self):
        results = Dataset.all()
        self.assertEqual(results.meta['current_page'], 1)
        self.assertEqual(results.current_page, 1)

    def test_to_list_returns_list(self):
        results = Dataset.all()
        data = results.to_list()
        self.assertIsInstance(data, list)
        self.assertIsInstance(data[0], list)



class ModelBase(object):
    def __init__(self, code, raw_data=None, **options):
        self.code = code
        self._raw_data = raw_data
        self.options = options

    def to_list(self):
        if isinstance(self.__get_raw_data__(), dict):
            return list(self.__get_raw_data__().values())
        return self.__get_raw_data__()

    def __getattr__(self, k):
        if k[0] == '_':
            raise AttributeError(k)
        elif k in self.__get_raw_data__():
            return self.__get_raw_data__()[k]
        else:
            raise AttributeError(k)

    def __getitem__(self, k):
        return self.__get_raw_data__()[k]

    def data_fields(self):
        return list(self.__get_raw_data__().keys())

    def __get_raw_data__(self):
        return self._raw_data



from inflection import parameterize
import dateutil.parser
import re
from six import string_types


class Util(object):
    @staticmethod
    def constructed_path(path, params={}):
        for key in list(params.copy().keys()):
            modified_path = path.replace(":%s" % key, str(params[key]))
            if modified_path != path:
                params.pop(key, None)
                path = modified_path
        return path

    # http://stackoverflow.com/questions/38987/how-can-i-merge-two-python-dictionaries-in-a-single-expression
    @staticmethod
    def merge_to_dicts(x, y):
        z = x.copy()
        z.update(y)
        return z

    @staticmethod
    def methodize(string):
        return parameterize(string.replace(r'.', '')).replace(r'-', '_')

    @staticmethod
    def merge_options(key, dic, **options):
        updated = dic
        # try to merge if options already has key, otherwise just assign
        if key in options:
            # respect values in options over hash
            updated = Util.merge_to_dicts(dic, options[key])
        options[key] = updated
        return options

    @staticmethod
    def convert_to_dates(dic_or_list):
        if isinstance(dic_or_list, dict):
            for k, v in list(dic_or_list.items()):
                dic_or_list[k] = Util.convert_to_dates(v)
        elif isinstance(dic_or_list, list):
            for idx, v in enumerate(dic_or_list):
                dic_or_list[idx] = Util.convert_to_dates(v)
        else:
            return Util.convert_to_date(dic_or_list)

        return dic_or_list

    @staticmethod
    def convert_to_date(value):
        if isinstance(value, string_types) and re.search(r'^\d{4}-\d{2}-\d{2}$', value):
            # convert to datetime.date
            return dateutil.parser.parse(value).date()
        elif isinstance(value, string_types) and re.search(r'^\d{4}-\d{2}-\d{2}T[\d:\.]+Z$', value):
            # convert to datetime.datetime, default timezone is utc
            return dateutil.parser.parse(value)
        else:
            return value

    @staticmethod
    def convert_options(request_type, **options):
        if request_type == 'get':
            return Util._convert_options_for_get_request(**options)
        elif request_type == 'post':
            return Util._convert_options_for_post_request(**options)
        else:
            raise Exception('Can only convert options for get or post requests')

    @staticmethod
    def _convert_options_for_get_request(**options):
        new_options = dict()
        if 'params' in options.keys():
            for key, value in options['params'].items():
                is_dict = False
                if isinstance(value, list):
                    key = key + '[]'
                else:
                    if isinstance(value, dict) and value != {}:
                        new_value = dict()
                        is_dict = True
                        old_key = key
                        for k, v in value.items():
                            key = key + '.' + k
                            if isinstance(v, list):
                                key = key + '[]'
                            new_value[key] = v
                            key = old_key

                if is_dict:
                    new_options.update(new_value)
                else:
                    new_options[key] = value
        return {'params': new_options}

    @staticmethod
    def _convert_options_for_post_request(**options):
        new_options = dict()
        if 'params' in options.keys():
            for key, value in options['params'].items():
                if isinstance(value, dict) and value != {}:
                    new_value = dict()
                    is_dict = True
                    old_key = key
                    for k, v in value.items():
                        key = key + '.' + k
                        new_value[key] = v
                        key = old_key
                else:
                    is_dict = False

                if is_dict:
                    new_options.update(new_value)
                else:
                    new_options[key] = value
        return {'json': new_options}

    @staticmethod
    def convert_to_columns_list(meta, type):
        columns = []
        for key in meta:
            columns.extend([key[type]])
        return columns

    @staticmethod
    def convert_column_names(meta):
        if meta is None:
            return []

        # Dataset API call
        if 'column_names' in meta.keys():
            the_list = [Util.methodize(x) for x in meta['column_names']]
            return list(the_list)
        # Datatable API call
        elif 'columns' in meta.keys():
            return list([Util.methodize(x) for x in meta['columns']])
        else:
            return []



class QuandlError(RuntimeError):
    GENERIC_ERROR_MESSAGE = 'Something went wrong. Please try again. \
If you continue to have problems, please contact us at connect@quandl.com.'

    def __init__(self, quandl_message=None, http_status=None, http_body=None, http_headers=None,
                 quandl_error_code=None, response_data=None):
        self.http_status = http_status
        self.http_body = http_body
        self.http_headers = http_headers if http_headers is not None else {}

        self.quandl_error_code = quandl_error_code
        self.quandl_message = quandl_message if quandl_message is not None \
            else self.GENERIC_ERROR_MESSAGE
        self.response_data = response_data

    def __str__(self):
        if self.http_status is None:
            status_string = ''
        else:
            status_string = "(Status %(http_status)s) " % {"http_status": self.http_status}
        if self.quandl_error_code is None:
            quandl_error_string = ''
        else:
            quandl_error_string = "(Quandl Error %(quandl_error_code)s) " % {
                "quandl_error_code": self.quandl_error_code}
        return "%(ss)s%(qes)s%(qm)s" % {
            "ss": status_string, "qes": quandl_error_string, "qm": self.quandl_message
        }


class AuthenticationError(QuandlError):
    pass


class InvalidRequestError(QuandlError):
    pass


class LimitExceededError(QuandlError):
    pass


class NotFoundError(QuandlError):
    pass


class ServiceUnavailableError(QuandlError):
    pass


class InternalServerError(QuandlError):
    pass


class ForbiddenError(QuandlError):
    pass


class InvalidDataError(QuandlError):
    pass


class ColumnNotFound(QuandlError):
    pass
